import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from torch.utils.data import DataLoader


class MoveAutoEncoder(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_embeddings: int = 1024,
        embedding_dim: int = 8,
        beta: float = 1e-3,
    ):
        super().__init__()

        self.encoder = nn.Sequential(
            nn.Linear(input_size, hidden_size),
            nn.LayerNorm(hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.LayerNorm(hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.LayerNorm(hidden_size),
        )
        self.embedding_dim = embedding_dim
        r = embedding_dim**-0.5
        self.beta = beta
        self.codebook = nn.Parameter(
            torch.randn(num_embeddings, embedding_dim).uniform_(-r, r)
        )
        self.decoder = nn.Sequential(
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.LayerNorm(hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.LayerNorm(hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, input_size),
        )

    def encode(self, x):
        return self.encoder(x)

    def quantize(self, z):
        B = z.shape[0]
        z_flat = z.view(B, -1, self.embedding_dim).flatten(0, 1)
        d = (
            torch.sum(z_flat**2, dim=1, keepdim=True)
            + torch.sum(self.codebook**2, dim=1)
            - 2 * torch.matmul(z_flat, self.codebook.t())
        )
        min_encoding_indices = torch.argmin(d, dim=1)
        z_q = torch.embedding(self.codebook, min_encoding_indices)
        z_q = z_q.view_as(z)

        loss = F.mse_loss(z_q.detach(), z) + self.beta * F.mse_loss(z_q, z.detach())

        z_q = z + (z_q - z).detach()

        return z_q, loss

    def decode(self, x):
        return self.decoder(x)

    def forward(self, x):
        z = self.encode(x)
        z_q, embedding_loss = self.quantize(z)
        x_hat = self.decode(z_q)
        return x_hat, embedding_loss


def encode(
    vectors: np.ndarray,
    input_size: int,
    hidden_size: int,
    batch_size: int = 2048,
    lr: float = 3e-4,
    num_epochs: int = 1000,
):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = MoveAutoEncoder(input_size, hidden_size).to(device)
    torch_vectors = torch.from_numpy(vectors.astype(float)).float()
    loader = DataLoader(torch_vectors, batch_size)
    optimizer = optim.Adam(model.parameters(), lr=lr)
    criterion = nn.MSELoss()

    while True:
        try:
            for batch in loader:
                optimizer.zero_grad()

                batch = batch.to(device)

                pred, loss = model(batch)
                loss += criterion(batch, pred)

                loss.backward()
                optimizer.step()

                print(f"{loss.item():.3f}", end="\r")
        except KeyboardInterrupt:
            break

        if loss.item() < 5e-3:
            break

    model = model.cpu()
    with torch.no_grad():
        z_q, _ = model.quantize(model.encode(torch_vectors))
        return z_q.cpu().detach().numpy()



import json
import os
import pickle
import torch
import torch.nn as nn


def load_gen_data(gen: int):
    with open(os.path.join(os.getcwd(), "pokemb/data/data.json"), "r") as f:
        return json.load(f)[gen]


def load_pkl_gen(gen: int):
    with open(os.path.join(os.getcwd(), "pokemb/data/encodings.pkl"), "rb") as f:
        return pickle.load(f)[gen]


class PokEmb(nn.Module):
    def __init__(self, gen: int = 9):
        super().__init__()

        raw = load_gen_data(f"gen{gen}")
        gendata = load_pkl_gen(f"gen{gen}")

        self.species = nn.Embedding.from_pretrained(
            torch.from_numpy(gendata["species"])
        )
        self.species_names = [v["id"] for v in raw["species"] if v["tier"] != "Illegal"]

        self.moves = nn.Embedding.from_pretrained(torch.from_numpy(gendata["moves"]))
        self.moves_names = [v["id"] for v in raw["moves"]]

        self.abilities = nn.Embedding.from_pretrained(
            torch.from_numpy(gendata["abilities"])
        )
        self.abilities_names = [v["id"] for v in raw["abilities"]]

        self.items = nn.Embedding.from_pretrained(torch.from_numpy(gendata["items"]))
        self.items_names = [v["id"] for v in raw["items"]]

        self.conditions = nn.Embedding.from_pretrained(
            torch.from_numpy(gendata["conditions"])
        )
        self.conditions_names = raw["conditions"]

        self.typechart = nn.Embedding.from_pretrained(
            torch.from_numpy(gendata["typechart"])
        )
        self.typechart_names = raw["typechart"]

    def forward(self, indices: torch.Tensor) -> torch.Tensor:
        ...

    def forward_species(self, indices: torch.Tensor) -> torch.Tensor:
        return self.species(indices)

    def forward_moves(self, indices: torch.Tensor) -> torch.Tensor:
        return self.moves(indices)

    def forward_abilities(self, indices: torch.Tensor) -> torch.Tensor:
        return self.abilities(indices)

    def forward_items(self, indices: torch.Tensor) -> torch.Tensor:
        return self.items(indices)

    def forward_conditions(self, indices: torch.Tensor) -> torch.Tensor:
        return self.conditions(indices)

    def forward_typechart(self, indices: torch.Tensor) -> torch.Tensor:
        return self.typechart(indices)



from setuptools import find_packages, setup

from pokemb import __version__

requirements = [
    "numpy==1.21.5",
    "pandas==2.0.2",
    "scikit_learn==1.3.2",
    "torch==1.13.0",
]

setup(
    name="pokemb",
    version=__version__,
    url="https://github.com/spktrm/pokemb",
    author="Joseph Twin",
    author_email="joseph.twin14@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    setup_requires=["wheel"],
    install_requires=requirements,
    package_data={"": ["*.ini"]},
)
